import numpy as np
import pandas as pd

from typing import Union, Optional
from functools import singledispatch

from tsp import TSP
from tspmetrics.classification import in_active_layer


""" Active layer thickness, top of permafrost, and thaw depth estimation """


def envelope_extrapolate(depths: Union[np.ndarray, pd.Series],
                         times: pd.DatetimeIndex,
                         values: np.ndarray, 
                         kind: str = 'alt') -> pd.Series:
    """ Extrapolate depth of 0 C isotherm using last two thawed nodes in a profile 
    using the maximum temperature at each depth over the year
    
    Parameters
    ----------
    depths : Union[np.ndarray, pd.Series]
        Depths of temperature observations
    times : pd.DatetimeIndex
        Times of temperature observations
    values : np.ndarray 
        Temperature observations
        
    Returns
    -------
    pd.Series
        The estimated depth of the 0 C isotherm at each time
    """
    df = pd.DataFrame(values, index=times, columns=depths)
    max_t = df.groupby(df.index.year).max()
    alt_max = max_t.apply(extrapolate_envelope_year, axis=1)
    min_t = -df.groupby(df.index.year).min()  # negative sign to use same function as alt_max
    alt_min = min_t.apply(extrapolate_envelope_year, axis=1)
    alt = pd.concat([alt_min,alt_max],axis=1).min(axis=1)  # shallowest intersection w/ 0C
    ttop = alt_max
    if kind == 'alt':
        return alt
    elif kind == 'ttop':
        return ttop
    else:
        raise ValueError("kind must be 'alt' or 'ttop'")


def envelope_interpolate(depths: Union[np.ndarray, pd.Series],
                         times: pd.DatetimeIndex,
                         values: np.ndarray,
                         kind='alt') -> pd.Series:
    """ Interpolate depth of 0 C isotherm between thawed and frozen nodes in a profile 
    using the maximum temperature at each depth over the year
    
    Parameters
    ----------
    depths : Union[np.ndarray, pd.Series]
        Depths of temperature observations
    times : pd.DatetimeIndex
        Times of temperature observations
    values : np.ndarray 
        Temperature observations
    kind : str
        either 'ttop' or 'alt', whether to return active layer thickness or top of permafrost
        
    Returns
    -------
    pd.Series
        The estimated depth of the 0 C isotherm at each time
    """
    df = pd.DataFrame(values, index=times, columns=depths)
    max_t = df.groupby(df.index.year).max()
    alt_max = max_t.apply(interpolate_envelope_year, axis=1)
    min_t = -df.groupby(df.index.year).min()  # negative sign to use same function as alt_max
    alt_min = min_t.apply(interpolate_envelope_year, axis=1)
    alt = pd.concat([alt_min,alt_max],axis=1).min(axis=1)  # shallowest intersection w/ 0C
    ttop = alt_max
    if kind == 'ttop':
        return ttop
    elif kind == 'alt':
        return alt
    else:
        raise ValueError("kind must be 'alt' or 'ttop'")


def interpolate_envelope_year(profile: pd.Series) -> pd.DataFrame:
    """ Interpolate depth of 0 C isotherm between thawed and frozen nodes in a profile
    
    Parameters
    ----------
    profile : pd.Series
        A profile of annual maximum temperatures with depth as the index

    Returns
    -------
    float
        The depth of the 0 C isotherm interpolated between the last thawed and first frozen nodes
    """
    W = np.where(profile < 0)
    if len(W[0]) == 0:
        return np.nan
    else:
        i_first_frozen = W[0][0]
    i_last_thawed = i_first_frozen - 1

    if i_last_thawed < 0:
        return np.nan

    d0 = profile.index[i_last_thawed]
    d1 = profile.index[i_first_frozen]
    t0 = profile[d0]
    t1 = profile[d1]

    dTdz =  (t1 - t0) / (d1 - d0) 

    d_zero = d0 + (0 - t0) / dTdz

    if np.sign(d_zero) != np.sign(d1):
        return np.nan
    
    return d_zero


def extrapolate_envelope_year(profile: pd.Series) -> pd.DataFrame:
    """ Extrapolate depth of 0C isotherm from last two thawed nodes in a profile
    Parameters
    ----------
    profile : pd.Series
        A profile of annual maximum temperatures with depth as the index

    Returns
    -------
    float
        The depth of the 0 C isotherm extrapolated using the last two thawed nodes
    """
    W = np.where(profile < 0)
    
    if len(W[0]) == 0:  # no frozen nodes
        return np.nan
    else:
        i_first_frozen = W[0][0]
    
    i_last_thawed = i_first_frozen - 1
    i_second_thawed = i_last_thawed - 1
    
    if i_second_thawed < 0:
        return np.nan
    
    d0 = profile.index[i_second_thawed]
    d1 = profile.index[i_last_thawed]
    t0 = profile[d0]
    t1 = profile[d1]
    
    dTdz =  (t1 - t0) / (d1 - d0) 
    
    d_zero = d1 + (0 - t1) / dTdz
    
    if np.sign(d_zero) != np.sign(d1):
        return np.nan
    
    return d_zero


def _thermal_gradients(depths, values) -> np.ndarray:
    return np.diff(values, axis=1) / np.diff(depths)


def deepest_al_node(depths, times, values, index=False, infer_start=False) -> pd.DataFrame:
    """ Determine deepest active layer node for each time step 
    
    np.nan if it can't find one
    """
    al = in_active_layer(depths, times, values, infer_start=infer_start).to_numpy()
    al[al != True] = False
    d = np.ma.array(np.repeat(np.atleast_2d(depths), repeats=al.shape[0], axis=0), mask=~al.astype('boolean'))
    deepest = np.apply_along_axis(lambda r: _get_deepest(r, depths, index=index), axis=1, arr=d)
    df = pd.DataFrame(data = deepest, index=times)
    return df


def _get_deepest(seq, depths, index):  
    if len(seq) == 0:  # possibly empty
        return np.nan
    else:
        i = np.argmax(seq)
        
        if index:
            return i
        
        return depths[i]  # an index


def deepest_thawed_al_node(depths, times, values, index=False, infer_start=False):
    """ The deepest thawed node in the active layer at each time  """
    al = in_active_layer(depths, times, values, infer_start=infer_start).to_numpy()
    al_thawed = al * (values > 0)
       
    deepest = np.apply_along_axis(lambda r: _get_deepest(np.where(r==1)[0], depths, index), axis=1, arr=al_thawed)
   
    return deepest


def shallowest_frozen_al_node(depths, times, values):
    """ The shallowest frozen node in the active layer at each time  """
    al = in_active_layer(depths, times, values).to_numpy()
    al_frozen = al * (values < 0)

    def get_shallowest(seq):  
        if len(seq) == 0:  # possibly empty
            return np.nan
        else:
            i = np.argmin(seq)
            return depths[i]  # an index
    
    shallowest = np.apply_along_axis(lambda r: get_shallowest(np.where(r==1)[0]), axis=1, arr=al_frozen)

    return shallowest


def instantaneous_zero_degree_isotherm(depths, times, values, method="extrapolate") -> pd.Series:
    """ Estimate the depth of the instantaneous zero-degree isotherm using either interpolation or extrapolation
    
    Parameters
    ----------
    depths : array-like
        Depths of the nodes in the profile
    times : pd.DatetimeIndex
        Times for each measurement
    values : array-like
        Temperature values for each node at each time
    method : str
        Method to use for estimating the zero-degree isotherm. Must be 'interpolate' or 'extrapolate'

    Returns
    -------
    pd.Series  
        Estimated depth of the instantaneous zero-degree isotherm for each time

    Extrapolation from the 2 thawed nodes above the zero degree isotherm is more accurate when there is 
    extensive phase change going on in permafrost (Riseborough, 2008)
    """
    dTdz = _thermal_gradients(depths, values)
    
    d = np.repeat(np.atleast_2d(depths[1:]), repeats=values.shape[0], axis=0)
    
    if method=='extrapolate':
        intercept = d + (0 - values[:, 1:]) / dTdz
    elif method=='interpolate':
        intercept = d + (0 - values[:, :-1]) / dTdz
    else:
        raise ValueError(f"Method {method} not recognized. Must be 'interpolate' or 'extrapolate'")

    df = pd.DataFrame(intercept, index=times, columns=depths[1:])

    # Identify the deepest positive-temperature observation within the AL
    o = deepest_thawed_al_node(depths, times, values, index=True, infer_start=True)
    o[o==0] = np.nan  # require 2 observations to extrapolate
    df['node'] = o

    freezeback_mask = shallowest_frozen_al_node(depths, times, values) < deepest_thawed_al_node(depths, times, values)
    df['freezeback'] = freezeback_mask

    def f(row):
        if row['freezeback']:  # possibly unreliable
            return np.nan
        if np.isnan(row['node']):  # no thawed nodes
            return np.nan
        else:
            i = int(row['node'])
            isotherm = row[depths[i]]

            if isotherm < depths[i]:
                return np.nan

            if (len(depths) > (i + 1)) and (isotherm > depths[i + 1]):
                return np.nan
        
        return isotherm
    
    result = df.apply(f, axis=1)
    result = result.groupby(result.index.year).max()

    return result


def zero_intercept(d1:float, d2:float, T1:float, T2:float, d3:Optional[float]=None) -> float:
    """ Calculate the zero-intercept of a thermal gradient
    
    @Parameters
    -----------
    d1 : float
        Depth at the first (uppermost) temperature measurement. Positive downwards.
    d2: float
        Depth at the second (second uppermost) temperature measurement. Positive downwards.
    T1 : float
        Temperature at d1
    T2 : float
        Temperature at d2
    d3 : float
        Depth at the third (lowermost) temperature measurement. Used as a check for the zero-intercept
    
    Returns
    -------
    float
        The depth at which the extrapolated thermal gradient crosses 0C. NaN if the intercept is below d2 or d3
    
    _____________________
    ----------|------o---
    ----------|-----/----
    -------d1-|----o-----
    ----------|---/------
    -------d2-|--o-------
    ----------|-/--------
    ----------|/---------
    ----------x----------
    ---------/|----------
    ---d3---o-|----------
    """
    dTdz =  (T2 - T1) / (d2 - d1) 
    zero_intercept = d2 + (0 - T2) / dTdz

    if zero_intercept < d2:
        return np.nan
    
    if d3 is not None:
        if zero_intercept > d3:
            zero_intercept = np.nan
    
    return zero_intercept


def top_of_permafrost(depths: Union[np.ndarray, pd.Series],
                      times: pd.DatetimeIndex,
                      values: np.ndarray,
                      estimation='extrapolate') -> pd.Series:
    """ Estimate the top of permafrost using the zero-degree isotherm 
    
    Parameters
    ----------
    depths : Union[np.ndarray, pd.Series]
        Depths of temperature observations
    times : pd.DatetimeIndex    
        Times of temperature observations
    values : np.ndarray
        Temperature observations
    estimation : str
        Method for estimating the top of permafrost. Options are 'extrapolate' and 'interpolate'

    Returns
    -------
    pd.Series
        The estimated top of permafrost at each time
    """
    if estimation == 'extrapolate':
        res = envelope_extrapolate(depths, times, values, kind='ttop')
    elif estimation == 'interpolate':
        res = envelope_interpolate(depths, times, values, kind='ttop')
    else:
        raise ValueError("estimation must be 'extrapolate' or 'interpolate'")
    
    return res


def active_layer_thickness(depths: Union[np.ndarray, pd.Series],
                      times: pd.DatetimeIndex,
                      values: np.ndarray,
                      estimation='extrapolate',
                      instantaneous=False) -> pd.Series:
    """ Estimate the active layer thickness using the zero-degree isotherm 
    
    Parameters
    ----------
    depths : Union[np.ndarray, pd.Series]
        Depths of temperature observations
    times : pd.DatetimeIndex    
        Times of temperature observations
    values : np.ndarray
        Temperature observations
    estimation : str
        Method for estimating the top of permafrost. Options are 'extrapolate' and 'interpolate'
    instantaneous : bool
        Whether to use instantaneous zero-degree isotherm estimation. Otherwise annual temperature envelope is used.

    Returns
    -------
    pd.Series
        The estimated top of permafrost at each time
    """
    if instantaneous:   
        res = instantaneous_zero_degree_isotherm(depths, times, values, method=estimation)
    else:
        if estimation == 'extrapolate':
            res = envelope_extrapolate(depths, times, values, kind='alt')
        elif estimation == 'interpolate':
            res = envelope_interpolate(depths, times, values, kind='alt')
        else:
            raise ValueError("estimation must be 'extrapolate' or 'interpolate'")
    
    return res


def tsp_top_of_permafrost(t:TSP, estimation='extrapolate') -> pd.Series:
    return top_of_permafrost(t.depths, t.times, t.values, estimation=estimation)


def tsp_active_layer_thickness(t:TSP, estimation='extrapolate', instantaneous=False) -> pd.Series:
    return active_layer_thickness(t.depths, t.times, t.values, estimation=estimation, instantaneous=instantaneous)



import pandas as pd
import numpy as np

from scipy.interpolate import interp1d
from tsp import TSP

from tspmetrics.amplitude import tsp_amplitude

""" Functions for calculating zero annual amplitude """


def dzaa_interpolated(amplitudes, zero_amplitude=0.1, dz=0.05):
    """ Calculate DZAA using (max-min)/2 as the amplitude.

    Parameters
    ----------
    amplitudes : pandas.DataFrame
        A DataFrame with depths as the columns and amplitudes as the values.
    zero_amplitude: float

    dz: float
        discretization interval for interpolating soil profile

    Returns
    -------
    pd.Series
        Series of depths of zero annual amplitude for each year in df.
    """
    def get_dzaa(S) -> float:
        ''' interpolate to get the amplitude at the zero_amplitude value '''
        f = interp1d(S.index, S.values)
        _z = np.arange(min(S.index), max(S.index), dz)
        _A = f(_z)
        dzaa = _z[np.argmin(np.abs(_A - zero_amplitude))]
        
        return dzaa
    
    return amplitudes.apply(get_dzaa, axis=1)


def dzaa_regressed(amplitudes: pd.DataFrame,
                   zero_amplitude=0.1,
                   cutoff=0.5) -> pd.Series:
    """Regresses the log of the amplitude ratios against the log of the depths.
    
    Parameteres
    ----------
    amplitudes : pandas.DataFrame
        A DataFrame with depths as the columns and amplitudes as the values.
    zero_amplitude : float
        The amplitude value at which we consider the amplitude to be zero.
    cutoff : float
        The minimum amplitude to be used in the regression.

    Returns
    -------
    pd.Series : Series of depths of zero annual amplitude for each year in df

    Method from Bonnaventure et al. 2015 'The ground thermal regime across the Mackenzie Valley Corridor'
    """
    return amplitudes.apply(lambda S: _regress_dzaa(S.index.values, S.values, zero_amplitude=zero_amplitude, cutoff=cutoff), axis=1)


def _regress_dzaa(depths, amplitudes, zero_amplitude=0.1, cutoff=0.5) -> float:
    """Regresses the log of the amplitude ratios against the log of the depths.

    Parameters
    ----------
    depths : array-like
        The depths of the amplitude measurements.
    amplitudes : array-like
        The amplitude at each depth.
    zero_amplitude : float
        The amplitude value at which we consider the amplitude to be zero.
    cutoff : float
        The minimum amplitude to be used in the regression.

    Returns
    -------
    float : Depth of zero annual amplitude. Depths always positive
    
    Method from Bonnaventure et al. 2015 'The ground thermal regime across the Mackenzie Valley Corridor'
    """
    if np.all(np.isnan(amplitudes)):
        return np.nan
    depths = np.abs(np.array(depths))  # require positive depths
    y = np.log(amplitudes[amplitudes > cutoff])
    x = depths[amplitudes > cutoff].astype('float64')  # depths column can be str sometimes
    x0, x1 = np.polynomial.polynomial.polyfit(x, y, 1)  # y = x0 + x1 * x
    zaa = (np.log(zero_amplitude) - x0) / x1
    
    return zaa


def tza(annual_means: pd.DataFrame, dzaa: pd.Series) -> pd.Series:
    """ Calculate the temperature at the depth of zero annual amplitude
    
    Parameters
    ----------
    annual_means: DataFrame
        Dataframe of annual temperature means with DatetimeIndex
    dzaa: Series
        Series of depths of zero annual amplitude for each year in annual_means
    """
    dzaa = np.abs(dzaa)
    depths = annual_means.columns.astype('float64')
    ix = dzaa.apply(func=lambda V: np.argmin(np.abs(V-depths)))
    t_dzaa = pd.Series(index=annual_means.index, data={"dzaa":np.nan}, name='dzaa')
    
    for i in range(ix.shape[0]):
        j = ix.iloc[i]
        if j == 0:
            continue
        t_dzaa.iloc[i] = annual_means.iloc[i,j]
    
    return t_dzaa

def get_dzaa_method(method: str):
    """ Return the function for calculating DZAA """
    
    methods = ['interpolated', 'regressed']
    
    if method not in methods:
        raise ValueError(f"method must be one of {methods}")
    elif method == 'interpolated':
        return dzaa_interpolated
    elif method == 'regressed':
        return dzaa_regressed
    else:
        raise ValueError(f"Unknown method {method}")
    

def dzaa(amplitudes, dzaa_method='interpolated', **kwargs) -> pd.Series:
    if dzaa_method == 'interpolated':
        return dzaa_interpolated(amplitudes, **kwargs)
    elif dzaa_method == 'regressed':
        return dzaa_regressed(amplitudes, **kwargs)
    else:
        raise ValueError(f"Unknown method {dzaa_method}")
    

def tsp_dzaa(t:TSP, dzaa_method='interpolated', amplitude_method='simple-all', **kwargs) -> pd.Series:
    """ Calculate the depth of zero annual amplitude for a TSP object
    
    Parameters
    ----------
    t : TSP
        A TSP object
    
    Returns
    -------
    pd.Series : Series of depths of zero annual amplitude for each year  
    """
    func_dzaa = get_dzaa_method(dzaa_method)
    amp = tsp_amplitude(t, method=amplitude_method)
    dzaa = func_dzaa(amp, **kwargs)
    
    return dzaa


def tsp_tza(t:TSP, dzaa_method='interpolated') -> pd.Series:
    """ Calculate the temperature at the depth of zero annual amplitude for a TSP object
    
    Parameters
    ----------
    t : TSP
        A TSP object
    
    Returns
    -------
    pd.Series : Series of temperatures at the depth of zero annual amplitude for each year
    """
    func_dzaa = get_dzaa_method(dzaa_method)
    amp = tsp_amplitude(t, method=amplitude_method)
    dzaa = func_dzaa(amp, **kwargs)

    t_dzaa = tza(t.yearly(), dzaa)
    
    return t_dzaa


from tsp import TSP
import pandas as pd
import numpy as np


from tspmetrics.labels import DATA_COMPLETENESS
from tspmetrics import labels as lbl


def in_active_layer(depths, times, values, infer_start=True) -> pd.DataFrame:
    """ Whether a depth node is in the active layer
    
    Parameters
    ----------
    depths : list
        List of depths
    times : pd.DatetimeIndex
        Time axis of the temperature values
    values : np.ndarray
        Temperature values at daily resolution or better
    infer_start : bool
        Infer status before first valid data

    Returns
    -------
    pd.DataFrame
        Dataframe of type 'boolean' with the same shape as the input data

    Description
    -----------
    Active layer is defined using a purely thermal definition, (e.g. ).
    """
    df = pd.DataFrame(values, index=times, columns=depths)
    has_pos = df.rolling("365D").max() > 0  # require temperatures above zero
    # 2-year for cryotic catches edge cases where ground might be below zero but partially thawed 
    # (and where there's no permafrost because it thawed) 
    has_neg = df.rolling("730D").min() < 0  # require temperatures below zero
    both = (has_pos & has_neg).astype('boolean')

    # missing data mask
    completeness = (~df.isna()).rolling("365D").sum() / 365 
    mdm = (completeness > DATA_COMPLETENESS).astype('boolean')
    both[~mdm] = pd.NA

    # stop AL classification when first non-AL, non-NA depth is encountered
    both = both.cummin(axis=1, skipna=True)

    # interpolate between missing datas 
    d_interp = boolean_interpolate_(both, axis=1)
    t_interp = boolean_interpolate_(both, axis=0)

    inferred = d_interp & t_interp

    # infer status before valid data
    if infer_start:
        for depth in both.columns:
            fvi = both[depth].first_valid_index()
            if fvi is not None:
                both.loc[:fvi, depth] = both.loc[fvi,depth]
                inferred.loc[:fvi, depth] = True

    return both


def in_talik(depths, times, values, inal=None, inpf=None):
    """ Determine which depth nodes are within a talik.

    """
    in_pf = inpf if inpf is not None else in_permafrost(depths, times, values)
    below_pf_table = in_pf.cummax(axis=1)
    in_al = inal if inal is not None else in_active_layer(depths, times, values)
    na_mask = in_al.isna() | below_pf_table.isna()
    df = pd.DataFrame(values, index=times, columns=depths)
    thawed_allyear = df.rolling("365D", min_periods=1).min() > 0
    talik = ~(in_al | below_pf_table) & thawed_allyear
    talik[na_mask] = pd.NA

    return talik


def in_permafrost(depths, times, values, infer_start=True) -> pd.DataFrame:
    """ Determine which depth nodes are within permafrost.
     Data must be daily averages 
    Returns
    -------
    pd.DataFrame
        Dataframe of type 'boolean' with the same shape as the input data
    """
    df = pd.DataFrame(values, index=times, columns=depths)
    cryotic_2years = (df.rolling("730D").max() < 0).astype("boolean")

    # missing data mask
    completeness = (~df.isna()).rolling("730D").sum() / 730 
    mdm = completeness > DATA_COMPLETENESS 
    cryotic_2years[~mdm] = pd.NA

    # 'sufficient time' mask (2 yr since first valid observation) 
    valid_obs = (~df.isna()).astype('int64')
    valid_times = valid_obs.apply(lambda col: col*(valid_obs.index.astype('int64')), axis=0)
    dt = valid_times.apply(lambda col: col - col[col.first_valid_index()]) * 1e-9
    adequate_span = dt.cummax(axis=1) > (730 * 24 * 60 * 60) 
    #import pdb;pdb.set_trace()
    cryotic_2years[~adequate_span] = pd.NA

    # interpolate between missing data
    inferred = boolean_bilinear_interpolate_(cryotic_2years)

    # infer status before valid data
    if infer_start:
        for depth in cryotic_2years.columns:
            fvi = cryotic_2years[depth].first_valid_index()
            if fvi is not None:
                cryotic_2years.loc[:fvi, depth] = cryotic_2years.loc[fvi,depth]
                inferred.loc[:fvi, depth] = True

    return cryotic_2years


def tsp_warm_pf(t: TSP) -> pd.DataFrame:
    """ Warm permafrost estimation """
    return in_warm_permafrost(t.depths, t.times, t.values, infer_start=True)


def tsp_classification(t: TSP, infer_start=False) -> pd.DataFrame:
    """ Classify a profile as active layer, permafrost, talik or unfrozen ground """
    is_pf = in_permafrost(t.depths, t.times, t.values, infer_start=infer_start)
    is_al = in_active_layer(t.depths, t.times, t.values, infer_start=infer_start)
    is_talik = in_talik(t.depths, t.times, t.values)

    classfication = is_pf.copy()
    classfication[is_pf.fillna(False)] = lbl.PERMAFROST
    classfication[is_al.fillna(False)] = lbl.ACTIVE_LAYER
    classfication[is_talik.fillna(False)] = lbl.TALIK

    classfication[~is_pf.fillna(False) & ~is_al.fillna(False) & ~is_talik.fillna(False)] = lbl.UNFROZEN
    classfication[is_pf.isna() & is_al.isna() & is_talik.isna()] = lbl.UNKNOWN

    # talik w/ no permafrost ? unfrozen
    uf = classfication.apply(lambda S: set(S.unique()) == set([lbl.ACTIVE_LAYER, lbl.TALIK]), axis=1)
    return classfication


## Ready
def boolean_interpolate_(df: pd.DataFrame, axis=0) -> pd.DataFrame:
    """ Interpolate IN PLACE between boolean values along a single axis

    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with boolean values
    axis : int
        Axis along which to interpolate (0 for rows, 1 for columns)
    
    Returns
    -------
    pd.DataFrame 
        Dataframe of indicating which values were interpolated (not the values themselves)

    Description
    -----------
    Will only interpolate beween values that are equal in both directions.
    """
    f = df.ffill(axis=axis)
    b = df.bfill(axis=axis)

    modified = (f==b) & (df.isna())

    df[f==b] = f[f==b]

    return modified


def boolean_bilinear_interpolate_(df: pd.DataFrame, strict=False) -> pd.DataFrame:
    """ Interpolate between boolean dataframe values IN PLACE considering both directions
    
    Parameters
    ----------
    df : pd.DataFrame
        Dataframe with boolean values and missing data (pandas.NA)
    strict : bool
        If True, only interpolate when both vertical and horizontal values are equal

    Returns
    -------
    pd.DataFrame
        boolean dataframe of which values were interpolated (not the values themselves)
    """
    if not df.dtypes.eq('boolean').all():
        raise ValueError("DataFrame must have boolean dtype for all columns")
    
    # Evaluate filling in all directions
    Vf = df.ffill(axis=0)
    Vb = df.bfill(axis=0)
    Hf = df.ffill(axis=1)
    Hb = df.bfill(axis=1)

    Veq = (Vf == Vb)  # verticals equal
    Heq = (Hf == Hb)  # horizontals equal
    
    is_na = df.isna()  # data we're interested in filling
    
    if strict:
        modified = (Veq & Heq) & (Vf == Hf) & is_na  # logically implies Vb == Hb
        df[modified] = Vf[modified]  # Vf could be anything here
    
    else:
        vf_isna = Vf.isna()
        hf_isna = Hf.isna()
        vb_isna = Vb.isna()
        hb_isna = Hb.isna()

        mod_V = Veq & ((Hf == Vf) | hf_isna) & ((Hb == Vf) | hb_isna) & is_na
        mod_H = Heq & ((Vf == Hf) | vf_isna) & ((Vb == Hf) | vb_isna) & is_na
        
        modified = mod_H | mod_V
        try:
            df[mod_V] = Vf[modified]
            df[mod_H] = Hf[modified]
        except:
            print("badbadbad")

    return modified


def in_warm_permafrost(depths, times, values, infer_start=False, cutoff=-2, inpf=None) -> pd.DataFrame:
    """ Whether permafrost is classified as warm for each depth and year"""
    permafrost = inpf if inpf is not None else in_permafrost(depths, times, values, infer_start=infer_start)
    permafrost = permafrost.resample("YE").min()
    df = pd.DataFrame(index=times, columns=depths, data=values)
    magt = df.resample("YE").mean()
    warm = (magt > cutoff).astype('boolean')
    warm_permafrost = warm & permafrost

    return warm_permafrost


def frozen_active_layer(depths, times, values):
    """ Whether active layer is completely frozen """
    al = in_active_layer(depths, times, values).to_numpy()
    al_thawed = al * (values > 0)

    return np.apply_along_axis(np.any, axis=1, arr=al_thawed).astype('boolean')



from alt import top_of_permafrost, active_layer_thickness, tsp_active_layer_thickness, tsp_top_of_permafrost
from amplitude import annual_amplitude, tsp_amplitude
from classification import in_active_layer, in_permafrost, in_talik, in_warm_permafrost, tsp_classification
from thaw_depth_duration import tsp_tdd, thaw_depth_duration
from thermal_integral import thermal_integral, tsp_thermal_integral
from zero_amplitude import dzaa, tza, tsp_tza, tsp_dzaa



import pandas as pd
import numpy as np

from tsp import TSP


def thermal_integral(depths, times, values, d0=None, dn=None):
    """ Thermal integral for each year 
    Parameters
    ----------
    depths : array-like
        Depths of the nodes in the profile
    times : pd.DatetimeIndex
        Times for each measurement
    values : array-like
        Temperature values for each node at each time
    T0 : float
        Temperature offset
    d0 : int
        Starting depth
    dn : int
        maximum depth
    """
    _ix = [i for i, d in enumerate(depths) if (d0 is None or d0 <= d) and (dn is None or d <= dn)]
    _depths = np.atleast_1d(depths)[_ix]
    _values = np.atleast_2d(values)[:, _ix]
    norm = max(_depths) - min(_depths)  # normalize to mean C/year
    df = pd.DataFrame(index=times, columns=_depths, data=_values)
    df = df.resample("YE").mean()
    # subset only permafrost depths?

    def _integrate(S):
        isnan = S.isna()
        if isnan.iloc[-1] or isnan.all():
            return np.nan
        else:
            clean = S.dropna()
            return np.trapezoid(clean.values, clean.index)
    # strict na policy
    #df = df.apply(lambda S: np.trapz(S.values, S.index), axis=1)
   
    # relaxed na policy
    df = df.apply(_integrate, axis=1)
    
    return df / norm


def tsp_thermal_integral(t:TSP, d0:float=0, dn):
    return thermal_integral(t.depths, t.times, t.values, d0=d0, dn=dn)


import numpy as np
import pandas as pd
from scipy.interpolate import InterpolatedUnivariateSpline as Interp
from functools import singledispatch

from tspmetrics.classification import in_permafrost
from tsp import TSP


def _tdd(depths, values, dz=0.01) -> float:
    """ Thaw depth duration increment for a single day 

    Parameters
    ----------
    depths : np.ndarray
        Depths of the temperature values
    values : np.ndarray
        Temperature values
    dz : float
        Depth increment for discretizing temperature profile
    
    Returns
    -------
    float
        Thaw depth duration increment for the day
    """
    T = Interp(depths, values, k=1, ext='const')
    z = np.arange(0, depths[-1], dz)
    Tz = T(z)
    thawed = np.heaviside(Tz, 0)
    
    return sum(thawed) * dz


@singledispatch
def thaw_depth_duration(depths: np.ndarray, times: "pd.DatetimeIndex", values:np.ndarray):
    """ 
    Calculate annual thaw depth duration

    Parameters
    ----------
    depths : np.ndarray
        Depths of the temperature values
    times : pd.DatetimeIndex
        Time axis of the temperature values
    values : np.ndarray
        Temperature values
    
    Returns
    -------
    pd.Series
        Thaw-depth duration for each year in the data frame

    Description
    -----------
    Based on Harp et al. (2016) doi:10.5194/tc-10-341-2016
    """
    # Ignore anything below permafrost table
    df = pd.DataFrame(index=times, columns=depths, data=values)
    in_pf = in_permafrost(depths,times,values)
    first_pf_node = in_pf.values.argmax(axis=1)
    df.insert(0, 'first_pf_node', first_pf_node)
    
    def skip_below(S):
        # Set temperatures below the top permafrost node to the same temperature as that node
        # This makes sure bottom taliks are not counted.
        S1 = S.iloc[1:]
        S1.iloc[int(S.iloc[0]):] = S1.iloc[int(S.iloc[0])]
        return S1

    cleaned = df.apply(skip_below, axis=1)

    r = cleaned.apply(lambda S: _tdd(S.ffill().index, S.ffill().values, name=S.name), axis=1)
    
    # If there's no permafrost, set to nan
    invalid = in_pf.apply(lambda S: ~S.any(), axis=1)
    r[invalid] = pd.NA
    # interpolate missing ffill
    Dbar = r.groupby(r.index.year).sum(min_count=365) / 365

    return Dbar


@thaw_depth_duration.register
def _(df: "pd.DataFrame"):
    depths = df.columns.astype('float64')
    times = df.index
    values = df.values
    return thaw_depth_duration(depths, times, values)


def tsp_tdd(t: TSP):
    """ Calculate thaw depth duration for a TSP object

    Parameters
    ----------
    t : TSP
        A TSP object
    
    Returns
    -------
    pd.Series
        Thaw-depth duration for each year in the data frame
    """
    return thaw_depth_duration(t.depths, t.times, t.values)


import pandas as pd
from enum import Enum
# classifications

TALIK = "talik"
PERMAFROST = "permafrost"
ACTIVE_LAYER = "active layer"
UNKNOWN = "unknown"
UNFROZEN = "unfrozen ground"

# Values 

DATA_COMPLETENESS = 0.90


# Resampling offsets
class YearDef(Enum):
    
    @staticmethod
    def list():
        return list(map(lambda c: c.value, YearDef))
    
    CALENDAR = "calendar_year"
    HYDRO = "hydrological_year"


def get_resample_offset(year=YearDef.CALENDAR):
    if year == YearDef.CALENDAR:
        return pd.offsets.YearEnd()
    elif year == YearDef.HYDRO:
        return pd.offsets.YearEnd(month=9)
    else:
        raise ValueError(f"year must be one of {YearDef.list()}")



import pandas as pd
from functools import singledispatch
from tsp import TSP

from tspmetrics.labels import YearDef, get_resample_offset
"""
Functions for calculating annual amplitude
"""


@singledispatch
def annual_amplitude(depths, times, values, method='simple-all', year_definition=YearDef.CALENDAR) -> pd.DataFrame:
    """ Calculate annual amplitude 
    Parameters
    ----------
    depths : list
        List of depths
    times : list
        List of times
    values : list
    method : str
        Method for calculating annual amplitude. Options are 'simple-all' (extrema using all values) 
        and 'simple-mm' (extrema of monthly means)
    year_definition : str
        Definition of a year. Valid options are provided in labels.YearDef

    Returns
    -------
    pd.DataFrame : DataFrame with annual amplitudes
    """
    valid_options = ['simple-all', 'simple-mm']
    if method == 'simple-all':
        A = simple_annual_maxmin(depths, times, values, year_definition=year_definition)
    elif method == 'simple-mm':
        A = monthly_mean_annual_maxmin(depths, times, values, year_definition=year_definition)
    else:
        raise ValueError(f'method must be one of {"".join(valid_options)}')
    
    return A


@annual_amplitude.register
def _(df: "pd.DataFrame", method='simple-all', year_definition=YearDef.CALENDAR) -> pd.DataFrame:
    depths = df.columns
    times = df.index
    values = df.values
    aa = annual_amplitude(depths,
                          times,
                          values, method=method,
                          year_definition=year_definition)
    
    return aa


def simple_annual_maxmin(depths, times, values, year_definition=YearDef.CALENDAR) -> pd.DataFrame:
    """ Amplitude over year, based on all available data """
    df = pd.DataFrame(index=pd.DatetimeIndex(times), columns=depths, data=values)
    df_annual = df.resample(get_resample_offset(year_definition))
    _range = df_annual.max() - df_annual.min()
    A = _range / 2

    return A



def monthly_mean_annual_maxmin(depths, times, values, year_definition=YearDef.CALENDAR) -> pd.DataFrame:
    """ Amplitude over year, based on monthly means """
    df = pd.DataFrame(index=times, columns=depths, data=values)
    monthly = df.resample("ME").mean()
    annual = df.resample(get_resample_offset(year_definition))
    _range = annual.max() - annual.min()
    A = _range / 2
    return A


def tsp_amplitude(t:TSP, method: str='simple-all', year_definition=YearDef.CALENDAR) -> pd.DataFrame:
    """ Calculate annual amplitude for a TSP object
    Parameters
    ----------
    t : TSP
        A TSP object
    method : str
        Method for calculating annual amplitude. Options are 'simple-all' and 'simple-mm'
    
    Returns
    -------
    pd.DataFrame : DataFrame with annual amplitudes
    """
    return annual_amplitude(t.depths, t.times, t.values, method, year_definition)


from setuptools import setup
import os
import re
from pathlib import Path


# Read version from VERSION file
base_dir = os.path.dirname(os.path.abspath(__file__))
version_path = os.path.join(base_dir, "VERSION")
try:
    with open(version_path) as version_file:
        version = version_file.read().strip()
except FileNotFoundError:  # can't find VERSION when building in isolated environment?
    expr = re.compile(r"tspmetrics-(\d+\.\d+(\.\d+)?)")
    vmatch = expr.match(Path(base_dir).name)
    if vmatch:
        version = vmatch.group(1)
    else:
        version = "0.0.0"

# Write version to __meta__.py (if needed)
meta_path = os.path.join(base_dir, "tspmetrics", "__meta__.py")
with open(meta_path, "w") as meta_file:
    data = (
f'''# Automatically created. Please do not edit.
__version__ = '{version}'
__author__ = 'Nick Brown'
'''
    )
    meta_file.write(data)

# Call the setup function
setup(version=version)